@@ -1,3 +1,15 @@
+# Copyright 2022 The Numaproj Authors.
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#     http://www.apache.org/licenses/LICENSE-2.0
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
+
 import logging
 from typing import Tuple
 
@@ -6,12 +18,9 @@ import torch.nn as nn
 from torch import Tensor
 from torch.nn.init import calculate_gain
 
-from numalogic.models.autoencoder.base import TorchAE
-from numalogic.preprocess.datasets import SequenceDataset
+from numalogic.models.autoencoder.base import BaseAE
 
 _LOGGER = logging.getLogger(__name__)
-_DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
-_LOGGER.info("Current device: %s", _DEVICE)
 
 
 class _Encoder(nn.Module):
@@ -80,7 +89,7 @@ class _Decoder(nn.Module):
         return out
 
 
-class LSTMAE(TorchAE):
+class LSTMAE(BaseAE):
     r"""
     Long Short-Term Memory (LSTM) based autoencoder.
 
@@ -90,7 +99,7 @@ class LSTMAE(TorchAE):
         embedding_dim: embedding dimension for the network
         encoder_layers: number of encoder layers (default = 1)
         decoder_layers: number of decoder layers (default = 1)
-
+        kwargs: BaseAE kwargs
     """
 
     def __init__(
@@ -100,8 +109,9 @@ class LSTMAE(TorchAE):
         embedding_dim: int,
         encoder_layers: int = 1,
         decoder_layers: int = 1,
+        **kwargs
     ):
-        super().__init__()
+        super().__init__(**kwargs)
 
         self.seq_len = seq_len
         self.no_features = no_features
@@ -113,7 +123,6 @@ class LSTMAE(TorchAE):
             embedding_size=self.embedding_dim,
             num_layers=encoder_layers,
         )
-        self.encoder = self.encoder.to(_DEVICE)
         self.encoder.apply(self.init_weights)
 
         self.decoder = _Decoder(
@@ -123,7 +132,6 @@ class LSTMAE(TorchAE):
             hidden_size=embedding_dim,
             num_layers=decoder_layers,
         )
-        self.decoder = self.decoder.to(_DEVICE)
         self.decoder.apply(self.init_weights)
 
     @staticmethod
@@ -138,22 +146,58 @@ class LSTMAE(TorchAE):
                 nn.init.xavier_normal_(param, gain=calculate_gain("tanh"))
 
     def forward(self, x: Tensor) -> Tuple[Tensor, Tensor]:
-        torch.manual_seed(0)
         encoded = self.encoder(x)
         decoded = self.decoder(encoded)
         return encoded, decoded
 
-    def construct_dataset(self, x: Tensor, seq_len: int = None) -> SequenceDataset:
-        r"""
-         Constructs dataset given tensor and seq_len
+    def predict_step(self, batch: Tensor, batch_idx: int, dataloader_idx: int = 0):
+        """Returns reconstruction for streaming input"""
+        recon = self.reconstruction(batch)
+        recon_err = self.criterion(batch, recon, reduction="none")
+        return recon_err
 
-         Args:
-            x: Tensor type
-            seq_len: sequence length / window length
+
+class SparseLSTMAE(LSTMAE):
+    r"""
+    Sparse Autoencoder for an LSTM network.
+    It inherits from VanillaAE class and serves as a wrapper around base network models.
+    Sparse Autoencoder is a type of autoencoder that applies sparsity constraint.
+    This helps in achieving information bottleneck even when the number of hidden units is huge.
+    It penalizes the loss function such that only some neurons are activated at a time.
+    This sparsity penalty helps in preventing overfitting.
+    More details about Sparse Autoencoder can be found at
+        <https://web.stanford.edu/class/cs294a/sparseAutoencoder.pdf>
+
+    Args:
+        beta: regularization parameter (Defaults to 1e-3)
+        rho: sparsity parameter value (Defaults to 0.05)
+        **kwargs: VanillaAE kwargs
+    """
+
+    def __init__(self, beta=1e-3, rho=0.05, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.beta = beta
+        self.rho = rho
+
+    def kl_divergence(self, activations: Tensor) -> Tensor:
+        r"""
+        Loss function for computing sparse penalty based on KL (Kullback-Leibler) Divergence.
+        KL Divergence measures the difference between two probability distributions.
+
+        Args:
+            activations: encoded output from the model layer-wise
 
         Returns:
-            SequenceDataset type
+            Tensor
         """
-        __seq_len = seq_len or self.seq_len
-        dataset = SequenceDataset(x, __seq_len, permute=False)
-        return dataset
+        rho_hat = torch.mean(activations, dim=0)
+        rho = torch.full(rho_hat.size(), self.rho, device=self.device)
+        kl_loss = nn.KLDivLoss(reduction="sum")
+        _dim = 0 if rho_hat.dim() == 1 else 1
+        return kl_loss(torch.log_softmax(rho_hat, dim=_dim), torch.softmax(rho, dim=_dim))
+
+    def _get_reconstruction_loss(self, batch):
+        latent, recon = self.forward(batch)
+        loss = self.criterion(batch, recon)
+        penalty = self.kl_divergence(latent)
+        return loss + penalty
